
#include "pm.h"
#include "small_chars.h"
#include "crc32.h"

#include <string.h>
#include <stdint.h>


#define BANK_SWITCH (*((volatile uint8_t _far *)0x1FFFFF))
#define FLASH_ADDR1 (*((volatile uint8_t _far *)0x1FFAAA))
#define FLASH_ADDR2 (*((volatile uint8_t _far *)0x1FF555))
#define SECTOR_ADDRESS 0x80000

#define BIOS_CRC32 0xAED3C14D

uint8_t params_in_ram[32];
uint8_t function_in_ram[1024];

void erase_sector(long address);
void write_byte(long address, uint8_t value);

static unsigned char reverse(unsigned char b) {
   b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
   b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
   b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
   return b;
}

static void switch_512k_0(void) {
   BANK_SWITCH = 0x80;
}

int main(void) {
   uint32_t i;

   // key interrupts priority
   PRI_KEY(0x03);
   //IRQ_PRI2 = 0x0c;
   // enable interrupts for D-pad keys
   IRQ_ENA3 = IRQ3_KEYPOWER;

   switch_512k_0();
   
   print(10, 0, "Calculating CRC...");
   
   i = calc_crc32(0, 0, 4096);
   if (i == BIOS_CRC32) {
      print(10, 1, "Standard BIOS!");
   } else {
      print(0, 0, " NON STANDARD BIOS!!!");
      print(0, 1, " NON STANDARD BIOS!!!");
      print(0, 2, " NON STANDARD BIOS!!!");
      print(0, 5, " NON STANDARD BIOS!!!");
      print(0, 6, " NON STANDARD BIOS!!!");
      print(0, 7, " NON STANDARD BIOS!!!");
   }

   if (memcmp((uint8_t _far *)0, (uint8_t _far *)0x80000, 4096) != 0) {
      print(10, 2, "Erasing sector...");
      erase_sector(SECTOR_ADDRESS);
      print(10, 3, "Writing address:");
      for (i = 0; i < 0x1000; ++i) {
         if((i&0xf) == 0)
            print_hex_long(10, 4, i);
         write_byte(SECTOR_ADDRESS + i, *((uint8_t *)i));
      }
      print(10, 6, "Done!");
   } else {
      print(4, 3, "The complete BIOS is");
      print(4, 4, "already dumped!");
   }

   for(;;)
      ;
}


static void run_in_ram(void (*f_begin)(void)) {   
   static uint8_t *previous = 0; 
   void (*f)(void);
   uint8_t *p1;

   p1 = (void *)f_begin;
   f = (void *)function_in_ram;
   if (p1 != previous) {
      memcpy(function_in_ram, p1, sizeof(function_in_ram));
      previous = p1;
   }

   (*f)();
}

void erase_sector_ram(void) {
   int i;
   unsigned long addr_long;
   FLASH_ADDR1 = 0x55;
   FLASH_ADDR2 = 0xAA;
   FLASH_ADDR1 = 0x01;
   FLASH_ADDR1 = 0x55;
   FLASH_ADDR2 = 0xAA;
   addr_long = *((long *)(&params_in_ram[1]));
   *((volatile uint8_t _far *)addr_long) = 0x0a;
   for (i = 0; i<1000; ++i) {
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
      _nop();
   }
}

void erase_sector(long address) {
   *((uint8_t **)(&params_in_ram[1])) = (uint8_t *)address;
   run_in_ram(erase_sector_ram);
}

void write_byte_ram(void) {
   unsigned long addr_long;
   FLASH_ADDR1 = 0x55;
   FLASH_ADDR2 = 0xAA;
   FLASH_ADDR1 = 0x05;
   addr_long = *((long *)(&params_in_ram[1]));
   *((volatile uint8_t _far *)addr_long) = params_in_ram[10];
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
   _nop();
}

void write_byte(long address, uint8_t value) {
   *((uint8_t **)(&params_in_ram[1])) = (uint8_t *)address;
   params_in_ram[10] = value;
   run_in_ram(write_byte_ram);
}

